/*
    Copyright 2008 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#ifdef WIN32
    #include <windows.h>
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.1.1"
#define PWD1        "Windows7621673NT"  // hardcoded
#define PWD2        "Virus3761267Trend" // hardcoded
#define PWDX        ""                  // hardcoded (0043d884) but empty, it's placed before the returned decrypted string



typedef struct {
    u8      *data;
    int     datasz;
    int     n;
} pwddecrypt_table_t;



void trendmicro_build_key(u8 *pwd1, u8 *pwd2, u8 *out);
void trendmicro_decrypt(u8 *string);
int hex2byte(u8 *input, u8 *output, int maxlen);
void wait_exit(int ret);
void std_err(void);



int main(int argc, char *argv[]) {
    FILE    *fd     = NULL;
    int     found   = 0;
    u8      buff[4096],
            *fname,
            *p;

    setbuf(stdout, NULL);

    fputs("\n"
        "TrendMicro passwords decrypter "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    if(argc < 2) {
        fprintf(stderr, "\n"
            "Usage: %s [file.ini/password]\n"
            "\n"
            "automatically decrypt any password in the input file or the encrypted string\n"
            "passed as first parameter.\n"
            "supports both !CRYPT! strings and those decriptable through the PWDDecrypt\n"
            "function of PWD.DLL.\n"
            "note that the !CRYPT! passwords are the encryptd MD5 hashes of the keywords.\n"
            "\n", argv[0]);
        wait_exit(1);
    }
    fname = argv[1];

    if(!strcmp(fname, "-")) {
        fd = stdout;
    } else {
        //fprintf(stderr, "- open file %s\n", fname);
        fd = fopen(fname, "rb");
    }
    if(!fd) {
        p = strchr(fname, '=');
        if(p) {
            p++;
        } else {
            p = fname;
        }
        trendmicro_decrypt(p);
        fprintf(stderr, "\n- decrypted string:\n\n");
        printf("%s\n", p);
        found++;
    } else {
        fprintf(stderr, "- loaded file %s\n\n", fname);
        while(fgets(buff, sizeof(buff), fd)) {
            for(p = buff; *p && (*p != '\r') && (*p != '\n'); p++);
            *p = 0;
            p = strchr(buff, '=');
            if(!p) continue;
            *p++ = 0;
            if(!*p) continue;
            if(stristr(buff, "pwd") || stristr(buff, "pass") || !strncmp(p, "!CRYPT!", 7)) {
                trendmicro_decrypt(p);
                printf("  %-20s %s\n", buff, p);
                found++;
            }
        }
        if(fd != stdout) fclose(fd);
    }

    if(!found) fprintf(stderr, "- no passwords found\n");
    wait_exit(0);
    return(0);
}



void trendmicro_build_key(u8 *pwd1, u8 *pwd2, u8 *out) {
    int     i,
            t;
    u16     x,
            *o;
    u8      *a,
            *b;

    for(i = 0; i < 8; i++) {
        out[i] = i + 1;
    }

    a = pwd1;
    b = pwd2;
    o = (u16 *)out;

    x = 0x7f3b;
    for(i = 0; i < 4; i++) {
        if(!*a && !*b) break;
        x ^= ((*a | (*b << 8)) ^ 0x6b2c);

        t = i;
        if(i == 1) {
            t = 2;
        } else if(i == 2) {
            t = 1;
        }
        o[t] ^= x;

        if(*a) a++;
        if(*b) b++;
    }
}



void trendmicro_crypt1(u8 *tablex, u8 *key) {
    static const u8 table1[] =  // 0042b364
        "\x07\x01\x06\x01\x05\x01\x04\x01\x03\x01\x02\x01\x01\x01\x00\x01"
        "\x07\x02\x06\x02\x05\x02\x04\x02\x03\x02\x02\x02\x01\x02\x00\x02"
        "\x07\x04\x06\x04\x05\x04\x04\x04\x03\x04\x02\x04\x01\x04\x00\x04"
        "\x07\x08\x06\x08\x05\x08\x04\x08\x07\x40\x06\x40\x05\x40\x04\x40"
        "\x03\x40\x02\x40\x01\x40\x00\x40\x07\x20\x06\x20\x05\x20\x04\x20"
        "\x03\x20\x02\x20\x01\x20\x00\x20\x07\x10\x06\x10\x05\x10\x04\x10"
        "\x03\x10\x02\x10\x01\x10\x00\x10\x03\x08\x02\x08\x01\x08\x00\x08";
    static const u8 table2[] =  // 0042b3d4
        "\x01\x01\x02\x02\x02\x02\x02\x02\x01\x02\x02\x02\x02\x02\x02\x01";
    static const u8 table3[] =  // 0042b3e4
        "\x0d\x10\x0a\x17\x00\x04\x02\x1b\x0e\x05\x14\x09\x16\x12\x0b\x03"
        "\x19\x07\x0f\x06\x1a\x13\x0c\x01\x28\x33\x1e\x24\x2e\x36\x1d\x27"
        "\x32\x2c\x20\x2f\x2b\x30\x26\x37\x21\x34\x2d\x29\x31\x23\x1c\x1f";

    int     i,
            j,
            c,
            r,
            t;
    u8      tmp[0x38];

    for(i = 0; i < 0x38; i++) {
        tmp[i] = (key[table1[i * 2]] & table1[(i * 2) + 1]) ? 1 : 0;
    }
    c = 0;
    for(i = 0; i < 0x10; i++) {
        c += table2[i];
        for(j = 0; j < 0x30; j++) {
            t = table3[j];
            if(t < 0x1c) {
                r = 0;
            } else {
                r = 0x1c;
                t -= 0x1c;
            }
            t += c;
            if(t >= 0x1c) t -= 0x1c;
            t += r;
            tablex[(i * 0x30) + j] = tmp[t];
        }
    }
}



void trendmicro_crypt2(u8 *tablex, u8 *data, int encdec) {
    static const u8 table1[] =  // 0042b2a4
        "\x07\x02\x06\x02\x05\x02\x04\x02\x03\x02\x02\x02\x01\x02\x00\x02"
        "\x07\x08\x06\x08\x05\x08\x04\x08\x03\x08\x02\x08\x01\x08\x00\x08"
        "\x07\x20\x06\x20\x05\x20\x04\x20\x03\x20\x02\x20\x01\x20\x00\x20"
        "\x07\x80\x06\x80\x05\x80\x04\x80\x03\x80\x02\x80\x01\x80\x00\x80"
        "\x07\x01\x06\x01\x05\x01\x04\x01\x03\x01\x02\x01\x01\x01\x00\x01"
        "\x07\x04\x06\x04\x05\x04\x04\x04\x03\x04\x02\x04\x01\x04\x00\x04"
        "\x07\x10\x06\x10\x05\x10\x04\x10\x03\x10\x02\x10\x01\x10\x00\x10"
        "\x07\x40\x06\x40\x05\x40\x04\x40\x03\x40\x02\x40\x01\x40\x00\x40";
    static const u8 table2[] =  // 0042b414
        "\x1f\x00\x01\x02\x03\x04\x03\x04\x05\x06\x07\x08\x07\x08\x09\x0a"
        "\x0b\x0c\x0b\x0c\x0d\x0e\x0f\x10\x0f\x10\x11\x12\x13\x14\x13\x14"
        "\x15\x16\x17\x18\x17\x18\x19\x1a\x1b\x1c\x1b\x1c\x1d\x1e\x1f\x00";
    static const u8 table3[] =  // 0042b444
        "\x0e\x04\x0d\x01\x02\x0f\x0b\x08\x03\x0a\x06\x0c\x05\x09\x00\x07"
        "\x00\x0f\x07\x04\x0e\x02\x0d\x01\x0a\x06\x0c\x0b\x09\x05\x03\x08"
        "\x04\x01\x0e\x08\x0d\x06\x02\x0b\x0f\x0c\x09\x07\x03\x0a\x05\x00"
        "\x0f\x0c\x08\x02\x04\x09\x01\x07\x05\x0b\x03\x0e\x0a\x00\x06\x0d"
        "\x0f\x01\x08\x0e\x06\x0b\x03\x04\x09\x07\x02\x0d\x0c\x00\x05\x0a"
        "\x03\x0d\x04\x07\x0f\x02\x08\x0e\x0c\x00\x01\x0a\x06\x09\x0b\x05"
        "\x00\x0e\x07\x0b\x0a\x04\x0d\x01\x05\x08\x0c\x06\x09\x03\x02\x0f"
        "\x0d\x08\x0a\x01\x03\x0f\x04\x02\x0b\x06\x07\x0c\x00\x05\x0e\x09"
        "\x0a\x00\x09\x0e\x06\x03\x0f\x05\x01\x0d\x0c\x07\x0b\x04\x02\x08"
        "\x0d\x07\x00\x09\x03\x04\x06\x0a\x02\x08\x05\x0e\x0c\x0b\x0f\x01"
        "\x0d\x06\x04\x09\x08\x0f\x03\x00\x0b\x01\x02\x0c\x05\x0a\x0e\x07"
        "\x01\x0a\x0d\x00\x06\x09\x08\x07\x04\x0f\x0e\x03\x0b\x05\x02\x0c"
        "\x07\x0d\x0e\x03\x00\x06\x09\x0a\x01\x02\x08\x05\x0b\x0c\x04\x0f"
        "\x0d\x08\x0b\x05\x06\x0f\x00\x03\x04\x07\x02\x0c\x01\x0a\x0e\x09"
        "\x0a\x06\x09\x00\x0c\x0b\x07\x0d\x0f\x01\x03\x0e\x05\x02\x08\x04"
        "\x03\x0f\x00\x06\x0a\x01\x0d\x08\x09\x04\x05\x0b\x0c\x07\x02\x0e"
        "\x02\x0c\x04\x01\x07\x0a\x0b\x06\x08\x05\x03\x0f\x0d\x00\x0e\x09"
        "\x0e\x0b\x02\x0c\x04\x07\x0d\x01\x05\x00\x0f\x0a\x03\x09\x08\x06"
        "\x04\x02\x01\x0b\x0a\x0d\x07\x08\x0f\x09\x0c\x05\x06\x03\x00\x0e"
        "\x0b\x08\x0c\x07\x01\x0e\x02\x0d\x06\x0f\x00\x09\x0a\x04\x05\x03"
        "\x0c\x01\x0a\x0f\x09\x02\x06\x08\x00\x0d\x03\x04\x0e\x07\x05\x0b"
        "\x0a\x0f\x04\x02\x07\x0c\x09\x05\x06\x01\x0d\x0e\x00\x0b\x03\x08"
        "\x09\x0e\x0f\x05\x02\x08\x0c\x03\x07\x00\x04\x0a\x01\x0d\x0b\x06"
        "\x04\x03\x02\x0c\x09\x05\x0f\x0a\x0b\x0e\x01\x07\x06\x00\x08\x0d"
        "\x04\x0b\x02\x0e\x0f\x00\x08\x0d\x03\x0c\x09\x07\x05\x0a\x06\x01"
        "\x0d\x00\x0b\x07\x04\x09\x01\x0a\x0e\x03\x05\x0c\x02\x0f\x08\x06"
        "\x01\x04\x0b\x0d\x0c\x03\x07\x0e\x0a\x0f\x06\x08\x00\x05\x09\x02"
        "\x06\x0b\x0d\x08\x01\x04\x0a\x07\x09\x05\x00\x0f\x0e\x02\x03\x0c"
        "\x0d\x02\x08\x04\x06\x0f\x0b\x01\x0a\x09\x03\x0e\x05\x00\x0c\x07"
        "\x01\x0f\x0d\x08\x0a\x03\x07\x04\x0c\x05\x06\x0b\x00\x0e\x09\x02"
        "\x07\x0b\x04\x01\x09\x0c\x0e\x02\x00\x06\x0a\x0d\x0f\x03\x05\x08"
        "\x02\x01\x0e\x07\x04\x0a\x08\x0d\x0f\x0c\x09\x00\x03\x05\x06\x0b";
    static const u8 table4[] =  // 0042b644
        "\x0f\x06\x13\x14\x1c\x0b\x1b\x10\x00\x0e\x16\x19\x04\x11\x1e\x09"
        "\x01\x07\x17\x0d\x1f\x1a\x02\x08\x12\x0c\x1d\x05\x15\x0a\x03\x18";
    static const u8 table5[] =  // 0042B324
        "\x27\x07\x2f\x0f\x37\x17\x3f\x1f\x26\x06\x2e\x0e\x36\x16\x3e\x1e"
        "\x25\x05\x2d\x0d\x35\x15\x3d\x1d\x24\x04\x2c\x0c\x34\x14\x3c\x1c"
        "\x23\x03\x2b\x0b\x33\x13\x3b\x1b\x22\x02\x2a\x0a\x32\x12\x3a\x1a"
        "\x21\x01\x29\x09\x31\x11\x39\x19\x20\x00\x28\x08\x30\x10\x38\x18";

    int     i,
            j,
            c,
            t;
    u8      tmp[0x60],  // ebp - 0x60
            tmp2[0x30], // ebp - 0x94
            bck[0x20];  // ebp - 0xc4

    for(i = 0; i < 0x40; i++) {
        tmp[i] = (data[table1[i * 2]] & table1[(i * 2) + 1]) ? 1 : 0;
    }
    for(i = 0; i < 0x10; i++) {
        memcpy(bck, tmp + 0x20, 0x20);
        if(!encdec) {
            t = i;
        } else {
            t = 0xf - i;
        }
        for(j = 0; j < 0x30; j++) {
            tmp2[j] = tmp[0x20 + table2[j]] ^ tablex[(t * 0x30) + j];
        }
        for(j = 0; j < 8; j++) {
            t = j * 6;
            c = (tmp2[t] << 5) | (tmp2[t + 1] << 3) | (tmp2[t + 2] << 2) | (tmp2[t + 3] << 1) | tmp2[t + 4] | (tmp2[t + 5] << 4);
            c = table3[(j << 6) + c];
            t = j << 2;
            tmp[0x40 + t] = (c >> 3) & 1;
            tmp[0x41 + t] = (c >> 2) & 1;
            tmp[0x42 + t] = (c >> 1) & 1;
            tmp[0x43 + t] = c & 1;
        }
        for(j = 0; j < 0x20; j++) {
            tmp[0x20 + j] = tmp[j] ^ tmp[0x40 + table4[j]];
        }
        memcpy(tmp, bck, 0x20);
    }

    memcpy(bck, tmp + 0x20, 0x20);
    memcpy(tmp + 0x20, tmp, 0x20);
    memcpy(tmp, bck, 0x20);

    t = 0;
    for(i = 0; i < 8; i++) {
        data[i] = 0;
        for(j = 1; j <= 0x80; j <<= 1) {
            if(tmp[table5[t]]) data[i] |= j;
            t++;
        }
    }
}



void trendmicro_PWDDecrypt(u8 *out, u8 *in, int *maxlen) {
    static const u8     key[]   = "199802232";
    static const int    tables  = 3;
    static const int    tablesz = 0x48;
    int     i,
            j,
            len;
    u8      tmp[tablesz],
            c;
    pwddecrypt_table_t  table[tables];

    c = 0;
    len = strlen(key);
    for(i = 0; i < tablesz; i++) {
        c += (key[i % len] + i) * 17;
        tmp[i] = key[(i + 1) % len] + (key[i % len] * c);
    }

    j = 0;
    len = (tablesz - tables) / tables;
    for(i = 0; i < tables; i++) {
        table[i].data   = tmp + j;
        table[i].datasz = len;
        table[i].n      = 0;
        j += len;
        len++;
    }

    len = hex2byte(in, out, -1);
    for(i = 0; i < len; i++) {
        c = 0;
        for(j = 0; j < tables; j++) {
            c ^= table[j].data[table[j].n];
            table[j].n++;
            if(table[j].n == table[j].datasz) table[j].n = 0;
        }
        out[i] ^= c;
    }
    out[i] = 0;
}



void trendmicro_decrypt(u8 *string) {
    u32     crc,
            mycrc;
    int     i,
            blocks,
            length;
    u8      tablex[768],
            key[8],
            ret[256],
            *str,
            *out,
            *p;

    if(!strncmp(string, "!CRYPT!", 7)) {            // !CRYPT!
        str = string + 7;
        out = string;
        // strcpy(ret, PWDX); this stupid and useless PWDX is no longer supported just because does nothing

        while(*str) {
            blocks = str[0] - '0';
            sscanf(str + 1, "%02X", &length);
            if((blocks > 8) || (length > 64)) {
                out[0] = 0;
                return;
            }
            str += 3;
            p = ret;

            trendmicro_build_key(PWD1, PWD2, key);

            trendmicro_crypt1(tablex, key);         // 0040a040
            for(i = 0; i < blocks; i++) {
                hex2byte(str, p, 8);                // 0040a910
                trendmicro_crypt2(tablex, p, 1);    // 0040a150
                str += 16;
                p   += 8;
            }
            ret[length] = 0;

            sscanf(str, "%08X", &mycrc);
            str += 8;

            p = ret;
            crc = 0x4f627a3b;
            for(i = 0; i < (length >> 2); i++) {
                crc ^= *(u32 *)p;
                p += 4;
            }
            if(crc != mycrc) {
                fprintf(stderr, "- alert: the CRC of the decrypted password (%08x) is wrong (%08x)\n", crc, mycrc);
            }
            strcpy(out, (out == string) ? (ret + 3) : ret);
            out += strlen(out);
        }
    } else {                                        // PWD.DLL
        fprintf(stderr, "- activated the PWDDecrypt method\n");

        length = strlen(string);                    // max size of the output buffer
        trendmicro_PWDDecrypt(string, string, &length);
    }
}



int hex2byte(u8 *input, u8 *output, int maxlen) {
    static const signed char hextable[256] =
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\xff\xff\xff\xff\xff\xff"
            "\xff\x0a\x0b\x0c\x0d\x0e\x0f\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\x0a\x0b\x0c\x0d\x0e\x0f\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
            "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff";
    u8      *i,
            *o;

    for(i = input, o = output; *i; o++) {
        if((maxlen >= 0) && ((o - output) >= maxlen)) break;
        while(*i && (hextable[*i] < 0)) i++;    // skipping of invalid chars
        if(!*i) break;
        *o = (hextable[*i] << 4);
        for(i++; *i && (hextable[*i] < 0); i++);
        if(!*i) break;
        *o |= hextable[*i];
        i++;
    }
    return(o - output);
}



void wait_exit(int ret) {
#ifdef WIN32
    if(GetWindowLong(GetForegroundWindow(), GWL_WNDPROC)) {
        fprintf(stderr, "\n  press RETURN to exit\n");
        fgetc(stdin);
    }
#endif
    exit(ret);
}



void std_err(void) {
    perror("\nError");
    wait_exit(1);
}


